/* $Id: isaac_wavelength.c,v 1.39 2013-03-12 08:06:48 llundin Exp $
 *
 * This file is part of the ISAAC Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA
 */

/*
 * $Author: llundin $
 * $Date: 2013-03-12 08:06:48 $
 * $Revision: 1.39 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#include "irplib_wavecal.h"
#include "irplib_utils.h"
#include "irplib_polynomial.h"
#include "isaac_wavelength.h"
#include "isaac_pfits.h"
#include "isaac_utils.h"
#include "isaac_dfs.h"

#include <string.h>
#include <math.h>
#include <float.h>

/*-----------------------------------------------------------------------------
                                   Define
 -----------------------------------------------------------------------------*/

#define ISAAC_MAX(A,B) ((A) > (B) ? (A) : (B))
#define ISAAC_MIN(A,B) ((A) < (B) ? (A) : (B))

#ifndef ISAAC_WFWHM
/* FIXME: Find a better value ? */
#define ISAAC_WFWHM 4.0
#endif

#define SLITWIDTH_TO_SIGMA 0.25

/* Evaluate 3rd degree (wavelength calibration) polynomial at
   ipix using Horners method
   Caveat: Both arguments are evaluated more than once here */
#define WAVELEN(poly,ipix) (poly[0] + (ipix) * \
                           (poly[1] + (ipix) * \
                           (poly[2] + (ipix) * \
                            poly[3])))

/* Evaluate differentiated 3rd degree (wavelength calibration) polynomial at
   ipix also using Horners method
   Caveat: Both arguments are evaluated more than once here */
#define WAVEDIF(poly,ipix) (poly[1]   + (ipix) * \
                           (poly[2]*2 + (ipix) * \
                            poly[3]*3))

/* Evaluate WAVELEN(poly, ipix + 0.5) - WAVELEN(poly, ipix - 0.5) in terms
   of WAVEDIF(). This is the width of pixel ipix in wavelengths
   Caveat: Both arguments are evaluated more than once here */
#define WAVEDLT(poly,ipix) (0.25*poly[3] + WAVEDIF(poly, ipix))

/* Default zone to discard in pixels */
#define ZEROPIX_LE                  10
#define ZEROPIX_RI                  10
/* Beginning of thermal regime in angstroms */
#define THERMAL_START               20000.0

/* Number of coefficients in the wavelength calibration polynomial
   - also the order of the resulting error term */
#define CALIB_COEFFS  4

/*----------------------------------------------------------------------------*/
/**
 * @defgroup isaac_wavelength       Wavelength calibration routines
 */
/*----------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------------
                            Function prototypes
 -----------------------------------------------------------------------------*/

static int isaac_wavelength_count_lines(const cpl_bivector *, double, double);
static int isaac_wavelength_count_linez(const cpl_bivector *, double, double);
static cpl_bivector * isaac_wavelength_load_catalog(const char *, const char *,
        const char *, const char *);
static isaac_band isaac_get_band(const char *);
static const char * isaac_get_filtername(isaac_band);
static double get_line_offset(const cpl_bivector *, const cpl_vector *,
        int, int, double, int, int, int, double, const double *, 
        double *, double *);
static double isaac_wavelength_centroid(const cpl_vector *, int, int);
static computed_disprel * spectro_refine_solution(const cpl_bivector *,
      const cpl_vector *, int, int, int, double, const double *);

/*-----------------------------------------------------------------------------
                                Function codes
 -----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief                  Compute a 3rd degree dispersion relation
  @param  in              Allocated spectroscopic image
  @param  discard_lo      Number of pixels to discard at the bottom
  @param  discard_hi      Number of pixels to discard at the top
  @param  discard_le      Number of pixels to discard on the left
  @param  discard_ri      Number of pixels to discard on the right
  @param  remove_thermal  Flag to force thermal background removal.
  @param  table_name      Spectral table name (see below)
  @param    oh          the oh lines catalog
  @param    ar          the ar lines catalog
  @param    xe          the xe lines catalog
  @param  slit_width      Width in pixels of the slit used
  @param  order           Order used in the spectral table look-up
  @param  output_ascii    Flag for ascii products
  @param  phdisprel       4 polynomial coefficients as first guess (phys. mod.?)
  @return a computed_disprel object containing among others the solution poly.

  Compute a dispersion relation from a spectroscopic image showing some
  strong emission lines. A vital assumption is that strong emission lines
  can be seen in the image.

  The first guess polynomial may have degree 2. In that case the 3rd degree 
  coefficient (4th value) must be zero.

  The removal of thermal background will attempt to remove from the
  spectroscopic image any low-frequency components, i.e. any features that
  are a lot wider than the slit width.
  
  The spectral table name is a character string. Possible values are:
  
  \begin{tabular}{ll}
  "oh"            &   OH lines \\
  "Xe"            &   Xenon lines \\
  "Ar"            &   Argon lines \\
  "Xe+Ar"         &   Xenon and Argon lines \\
  \end{tabular}
  
  The order determines at which location in the spectral table the look-up
  for emission lines will be done.

  The output cross-correlation factor is a number between 0 and 1. A good
  fit typically results in a value in the range 0.7 to 0.95, while a
  spectroscopic image without a number of well separated strong emission
  lines typically will yield a value down to about 0.5. Early spectroscopic
  images from the ESO ISAAC instrument cross-correlate down to 0.38.
  
  Algorithm:
  \begin{itemize}
  \item Spectrum extraction along the spectrum direction (horizontal).
  For each column the lower and higher
  pixels are discarded, then a median value of the remaining pixels in the 
  column is returned. This forms a 1d signal of same size as the image in the 
  spectrum direction.
  \item If a thermal background should be removed, this is done at this point.
  \item If some values must be discarded (set to zero) in the input spectrum, 
  they are zeroed at that point.
  \item Spectral lines are retrieved from a catalog 
  \item A search is done for the polynomial that maximizes the
  cross-correlation of the extracted signal and the spectral lines catalog.
  \end{itemize}

  Setting parts of the input spectrum to zero enables a correct wavelength 
  calibration in the thermal regime. You should provide -1 and -1 if you want 
  to let this function decide for you about a correct zeroing interval, (0,0) 
  if you do not want to zero the spectrum at all, and any other values 
  depending on which interval you want to set to zero on each side of the 
  spectrum.
  Notice that setting these values to (10,20) will set to zero the 10 left 
  pixels and 20 rightmost pixels of the input spectrum before throwing it into 
  the cross-correlation procedure.
*/
/*----------------------------------------------------------------------------*/
computed_disprel * spectro_compute_disprel(
        const cpl_image *   in,
        int                 discard_lo,
        int                 discard_hi,
        int                 discard_le,
        int                 discard_ri,
        int                 max_offset,
        int                 remove_thermal,
        const char      *   table_name,
        const char      *   oh,
        const char      *   ar,
        const char      *   xe,
        double              slit_width,
        int                 order,
        int                 output_ascii,
        const double    *   phdisprel)
{
    cpl_bivector            *   spec_cat;
    computed_disprel        *   solution;
    double                  *   disprel;
    cpl_image               *   collapsed;
    cpl_image               *   thresholded;
    cpl_vector              *   spec_init;
    cpl_vector              *   spec_bg;
    cpl_vector              *   spec_bgclean;
    double                  *   pspec_bgclean;
    cpl_vector              *   spec_bgclean_log;
    double                  *   pspec_bgclean_log;
    double                      xc = 0.0;
    const int                   npix = cpl_image_get_size_x(in);
    double                      wl_min, wl_max;
    int                         emil;
    cpl_size                    i;

    cpl_ensure(in         != NULL, CPL_ERROR_NULL_INPUT, NULL);
    cpl_ensure(table_name != NULL, CPL_ERROR_NULL_INPUT, NULL);
    cpl_ensure(phdisprel  != NULL, CPL_ERROR_NULL_INPUT, NULL);


    /* Initialize */
    wl_min = WAVELEN(phdisprel, 0.5);
    wl_max = WAVELEN(phdisprel, npix+0.5);

    /* Check acceptable wavelength range */
    if ((wl_min > wl_max) || (wl_min < MIN_WAVELENGTH) ||
        (wl_max > MAX_WAVELENGTH)) {
        cpl_msg_error(cpl_func, 
                "in provided wavelength range: [%g %g] ([min max] is [%g %g])",
                wl_min, wl_max, MIN_WAVELENGTH, MAX_WAVELENGTH);
        return NULL;
    }

    /* Load the catalog  */
    cpl_msg_info(cpl_func, "Load the catalog");
    /* FIXME: Multiply catalogue wavelengths with otder */
    if ((spec_cat=isaac_wavelength_load_catalog(table_name, oh,ar,xe))==NULL) {
        cpl_msg_error(cpl_func, "Cannot load the catalog");
        return NULL;
    }

    if (order > 1) {
        cpl_vector_multiply_scalar(cpl_bivector_get_x(spec_cat), order);
    }
    
    cpl_msg_info(cpl_func, "Cross correlation");
    /* Get the number of lines in the catalog for the specified range */
    emil = isaac_wavelength_count_lines(spec_cat, wl_min, wl_max);

    cpl_msg_debug(cpl_func, "Spectral order: %d", order);
    cpl_msg_debug(cpl_func, "First guess poly. wl = f(pix) (pix in 1-%d):", 
            npix);
    cpl_msg_debug(cpl_func, "f(x) = %g + %g * x + %g * x^2  + %g * x^3\n", 
            phdisprel[0], phdisprel[1], phdisprel[2], phdisprel[3]);
    cpl_msg_debug(cpl_func, "Spectral range [%g %g] with %02d out of %d lines",
                  wl_min, wl_max, emil, (int)cpl_bivector_get_size(spec_cat));

    /* Test if some lines are found in the catalog */
    if (emil < 1) {
        cpl_msg_error(cpl_func, 
                "No line found in catalog in the specified range - abort"); 
        cpl_bivector_delete(spec_cat);
        return NULL;
    }

    /* Threshold the image to remove negative values */
    thresholded = cpl_image_duplicate(in);
    if (cpl_image_threshold(thresholded, 0.0, DBL_MAX, 0.0, 0.0)!=CPL_ERROR_NONE) {
        cpl_msg_error(cpl_func, 
                "thresholding input image: aborting wavelength calibration");
        cpl_bivector_delete(spec_cat);
        cpl_image_delete(thresholded);
        return NULL;
    }

    /* Get a list of lines in the image by median-collapsing it over */
    /* the horizontal direction. With a little help from the median, */
    /* should get rid of cosmics and other spiky defects. */
    if ((collapsed = cpl_image_collapse_median_create(thresholded, 0, 
                    discard_lo, discard_hi)) == NULL) {
        cpl_msg_error(cpl_func, 
                "collapsing input image: aborting wavelength calibration");
        cpl_bivector_delete(spec_cat);
        cpl_image_delete(thresholded);
        return NULL;
    }
    cpl_image_delete(thresholded);

    /* Put the spectrum in a vector */
    spec_init = cpl_vector_new_from_image_row(collapsed, 1);
    cpl_image_delete(collapsed);

    /* Remove thermal background contributions above THERMAL_START */
    spec_bgclean = cpl_vector_duplicate(spec_init);
    if ((remove_thermal!=0&&wl_max>THERMAL_START) || !strcmp(table_name,"oh")){
        cpl_msg_debug(cpl_func, "Removing low-frequency background");
        /* Filter away very wide (8 * slit_width) features */
        if ((spec_bg = cpl_vector_filter_median_create(spec_init, 
                        (int)((0.5+8*slit_width)/2))) == NULL) {
            cpl_msg_error(cpl_func, "sub_lowpass failed");
            cpl_bivector_delete(spec_cat);
            cpl_vector_delete(spec_init);
            cpl_vector_delete(spec_bgclean);
            return NULL;
        }
        cpl_vector_subtract(spec_bgclean, spec_bg);
        cpl_vector_delete(spec_bg);
        /* Threshold negative intensities */
        pspec_bgclean = cpl_vector_get_data(spec_bgclean);
        for (i=0; i<cpl_vector_get_size(spec_bgclean); i++) 
            if (pspec_bgclean[i] < 0) pspec_bgclean[i] = 0.0;
    } 

    /* See if default zeroing widths have been requested */
    if (discard_le < 0 || discard_le >= npix) discard_le = ZEROPIX_LE;
    if (discard_ri < 0 || discard_le + discard_ri >= npix) {
        if (!strcmp(table_name, "oh") && wl_max > THERMAL_START) {
            /* Calibrate using OH lines */
            /* Thermal regime: discard the right side of the signal */
            discard_ri = npix/2;
        } else {
            /* Calibrate using standard lamps */
            discard_ri = ZEROPIX_RI;
        }
    }

    /* Put less weight on the intensity by taking the logarithm */
    /* Add 1 to ensure continuity around zero */
    spec_bgclean_log = cpl_vector_duplicate(spec_bgclean); 
    pspec_bgclean = cpl_vector_get_data(spec_bgclean);
    pspec_bgclean_log = cpl_vector_get_data(spec_bgclean_log);
    for (i=0; i<cpl_vector_get_size(spec_bgclean_log); i++)
      pspec_bgclean_log[i] = pspec_bgclean[i] > 0.0 ? 
          log1p(pspec_bgclean[i]) : 0.0;

    disprel = cpl_malloc(CALIB_COEFFS * sizeof(double));

    if (1) {  /* Start of irplib_wavecal based calibration */
        irplib_line_spectrum_model model;

        const double wfwhm = ISAAC_WFWHM;
        const double slitw = slit_width;
        const double xtrunc = 0.5 * slitw + 5.0 * wfwhm * CPL_MATH_SIG_FWHM;

        /* Extract the interesting part of the spectrum */
        cpl_vector * spec_short = cpl_vector_extract(spec_bgclean_log,
                                                     discard_le,
                                                     npix-discard_ri-1, 1);

        const int short_size = cpl_vector_get_size(spec_short);

        /* One practical limitation on hshiftmax is that it may not be so big,
           as to cause phdisp to be evaluated outside its range of monotony. */
        const int    hshiftmax = ISAAC_MIN(short_size/4, max_offset);

        /* Initialize to zero */
        cpl_vector     * linepix
            = cpl_vector_wrap(cpl_bivector_get_size(spec_cat),
                              cpl_calloc(cpl_bivector_get_size(spec_cat),
                                         sizeof(double)));
        cpl_vector     * erftmp
            = cpl_vector_wrap(1, cpl_calloc(1, sizeof(double)));
        cpl_polynomial * phshift = cpl_polynomial_new(1);
        cpl_error_code   error = CPL_ERROR_NONE;
        const int        fitdeg = emil >= CALIB_COEFFS
            ? CALIB_COEFFS-1 : emil-1;
        double           pixstep = 0.5;
        double           pixtol = 1e-5;
        const int        maxite = fitdeg * 200;
        int              maxfail = 3;
        int              maxcont = 1;
        const unsigned   clines
            = (unsigned)(xtrunc*(short_size + 2 * hshiftmax));
        const cpl_boolean doplot /* FIXME: Function parameter */
            = getenv("ISAAC_DOPLOT") ? CPL_TRUE : CPL_FALSE;

        memset(&model, 0, sizeof(model));
        model.wslit = slitw;
        model.wfwhm = wfwhm;
        model.xtrunc = xtrunc;
        model.lines = spec_cat;
        model.linepix = linepix;
        model.erftmp = erftmp;
        model.cost = 0;
        model.xcost = 0;

        for (i = 0; i < CALIB_COEFFS; i++) { /*  Revert order for efficiency */
            error |= cpl_polynomial_set_coeff(phshift, &i, phdisprel[i]);
        }
        error |= cpl_polynomial_shift_1d(phshift, 0, discard_le);

        error |= irplib_polynomial_find_1d_from_correlation_all
            (phshift, fitdeg, spec_short, 0, clines,
             (irplib_base_spectrum_model*)&model,
             irplib_vector_fill_logline_spectrum_fast, pixtol, pixstep,
             hshiftmax, maxite, maxfail, maxcont, doplot, &xc);

        error |= irplib_polynomial_find_1d_from_correlation_all
            (phshift, fitdeg, spec_short, 1, clines,
             (irplib_base_spectrum_model*)&model,
             irplib_vector_fill_logline_spectrum, pixtol, pixstep,
             hshiftmax, maxite, maxfail, maxcont, doplot, &xc);

        cpl_vector_delete(spec_short);
        cpl_vector_delete(linepix);
        cpl_vector_delete(erftmp);

        error |= cpl_polynomial_shift_1d(phshift, 0, -discard_le);

        if (!error) {
            cpl_msg_info(cpl_func, "XC: %g (cost=%u:%u. lines=%u)", xc,
                         (unsigned)model.cost, (unsigned)model.xcost,
                         (unsigned)model.ulines);

            for (i = 0; i < CALIB_COEFFS; i++) {
                disprel[i] = cpl_polynomial_get_coeff(phshift, &i);
            }
        } else {
            cpl_error_set_where(cpl_func);
            for (i = 0; i < CALIB_COEFFS; i++) {
                disprel[i] = phdisprel[i];
            }
        }

        cpl_polynomial_delete(phshift);

    } /* End of irplib_wavecal based calibration */

    if ((discard_le>0) || (discard_ri>0)) {
        double wl_min_z = wl_min;
        double wl_max_z = wl_max;
        int emil_z;
        if (discard_le>0) wl_min_z = WAVELEN(phdisprel, discard_le+0.5);
        if (discard_ri>0) wl_max_z = WAVELEN(phdisprel, npix-discard_ri+0.5);

        emil_z = isaac_wavelength_count_lines(spec_cat, wl_min_z, wl_max_z);

        cpl_msg_debug(cpl_func,
                "Zeroed calibration signal [%g %g] has %d lines, dropped %d",
                wl_min_z, wl_max_z, emil_z, emil-emil_z);
    }

    for (i=0; i<CALIB_COEFFS; i++)
        cpl_msg_debug(cpl_func, "Coef nb. %d correction rate: %g / %g = %g", 
                      (int)i+1, disprel[i], phdisprel[i], phdisprel[i] != 0 ?
                disprel[i]/ phdisprel[i]  : disprel[i]);

    /* Narrow refinable range by HALF_CENTROID_DOMAIN */
    solution = spectro_refine_solution(spec_cat, spec_bgclean, discard_le+5, 
            discard_ri-5, npix, slit_width, disprel);
    if (solution == NULL) {
        cpl_free(disprel);
        cpl_bivector_delete(spec_cat);
        cpl_vector_delete(spec_init);
        cpl_vector_delete(spec_bgclean);
        cpl_vector_delete(spec_bgclean_log);
        return NULL;
    }

    /* Produce some ASCII files with the computed signals */
    if (output_ascii) {
        FILE * out_file;
        /* First guess solution */
        out_file = fopen("collapsed_physical.txt", "w");
        for (i=0; i<npix; i++)
            fprintf(out_file, "%g\t%g\n", WAVELEN(phdisprel, i+1), 
                    cpl_vector_get(spec_init, i));
        fclose(out_file);

        /* Computed solution */
        out_file = fopen("collapsed_calibrated.txt", "w");
        for (i=0; i<npix; i++)
            fprintf(out_file, "%g\t%g\n", WAVELEN(disprel, i+1), 
                    cpl_vector_get(spec_init, i));
        fclose(out_file);

        /* Computed solution with the low frequency part removed */
        out_file = fopen("collapsed_calibrated_filtered.txt", "w");
        for (i=0; i<npix; i++)
            fprintf(out_file, "%g\t%g\n", WAVELEN(disprel, i+1), 
                    cpl_vector_get(spec_bgclean, i));
        fclose(out_file);
    }

    /* Free memory */
    cpl_vector_delete(spec_init);
    cpl_vector_delete(spec_bgclean);
    cpl_vector_delete(spec_bgclean_log);
    cpl_bivector_delete(spec_cat);

    /* Display results */
    cpl_msg_debug(cpl_func, "Computed poly. wave = f(pix) (pix in 1-%d):",npix);
    cpl_msg_debug(cpl_func, "f(x) = %g + %g * x + %g * x^2  + %g * x^3",
            disprel[0], disprel[1], disprel[2], disprel[3]);
    cpl_msg_debug(cpl_func, "Spectral range [%g %g]",
            WAVELEN(disprel,1), WAVELEN(disprel,npix));

    /* Fill in the solution and remaining parameters */
    solution->poly   = disprel;
    solution->degree = CALIB_COEFFS-1;
    solution->cc     = xc;
    solution->offset = (disprel[0] - phdisprel[0]) / 
	               WAVEDLT(disprel, 0.5 * (1 + npix));
    solution->scal1  = disprel[1] / phdisprel[1];
    solution->scal2  = disprel[2];
    if (phdisprel[2] != 0) solution->scal2 /= phdisprel[2];
    solution->scal3  = disprel[3];
    if (phdisprel[3] != 0) solution->scal3 /= phdisprel[3];
    disprel = NULL;

    /* Return the solution */
    return solution;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Determine the order according to filter, grating and wl
  @param    image_name  input image name
  @return   order, -1 in error case 
 */
/*----------------------------------------------------------------------------*/
int isaac_find_order(const char * image_name)
{
    int                     order;
    const char          *   sval;
    cpl_propertylist    *   plist;
    char                    grat_name[128];
    double                  wl_c;
    isaac_band              band;

    /* Initialize    */
    order = 1;
    plist=cpl_propertylist_load(image_name, 0);

    /* Get the grating name */
    if ((sval = isaac_pfits_get_resolution(plist)) == NULL) {
        cpl_msg_error(cpl_func, "cannot get resolution from [%s]", image_name);
        cpl_propertylist_delete(plist);
        return -1;
    }
    strcpy(grat_name, sval);
    /* Get the central wavelength */
    wl_c = isaac_pfits_get_wlen(plist);
    if (cpl_error_get_code()) {
        cpl_msg_error(cpl_func, "cannot get central wlen from [%s]", image_name);
        cpl_propertylist_delete(plist);
        return -1;
    }
    wl_c *= 10000.0; /* microns to A */
    /* Get the filter used */
    if ((sval = isaac_pfits_get_filter(plist)) == NULL) {
        cpl_msg_error(cpl_func, "cannot get filter from [%s]", image_name);
        cpl_propertylist_delete(plist);
        return -1;
    }
    band = isaac_get_band(sval);
    cpl_propertylist_delete(plist);

    /* Association rules - Medium resolution */
    if (grat_name[0] == 'M' && band == ISAAC_BAND_SH &&
        27000 < wl_c && wl_c < 42000) order = 2;

    if (grat_name[0] == 'M' && band == ISAAC_BAND_JBLOCK &&
        35500 < wl_c && wl_c < 42000) order = 3;

    /* This association is currently only relevant for historical data */
    if (grat_name[0] == 'M' && band == ISAAC_BAND_SK &&
        44000 < wl_c && wl_c < 51000) order = 2;

    if (grat_name[0] == 'M' && band == ISAAC_BAND_SH &&
        44000 < wl_c && wl_c < 51000) order = 3;

    /* This association is currently only relevant for historical data */
    if (grat_name[0] == 'M' && band == ISAAC_BAND_JBLOCK &&
        44000 < wl_c && wl_c < 51000) order = 4;

    /* Association rules - Low resolution */
#if 1
    /* verify with DFO */
    /* This association is currently only relevant for historical data */
    if (grat_name[0] == 'L' && band == ISAAC_BAND_SK &&
        35500 < wl_c && wl_c < 42000) order = 2;

    /* This association is currently only relevant for historical data */
    if (grat_name[0] == 'L' && band == ISAAC_BAND_SH &&
        35500 < wl_c && wl_c < 42000) order = 2;

    /* This association is currently only relevant for historical data */
     if (grat_name[0] == 'L' && band == ISAAC_BAND_JBLOCK &&
        35500 < wl_c && wl_c < 42000) order = 3;
#endif

    if (grat_name[0] == 'L' && band == ISAAC_BAND_SH &&
        44000 < wl_c && wl_c < 51000) order = 3;

    cpl_msg_debug(cpl_func, 
            "Find order: %d. Resol: %c. Lambda_c: %g. Filter: %s",
            order, grat_name[0], wl_c, isaac_get_filtername(band));
    return order;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Determine presence of thermal background according to instr. setting
  @param    image_name  input image name
  @return   1 if yes, 0, when not, -1 in error case 
 */
/*----------------------------------------------------------------------------*/
int isaac_has_thermal(const char * im_name)
{
    const char          *   sval;
    cpl_propertylist    *   plist;
    char                    grat_name[128];
    double                  wl_c;
    isaac_band              band;
    int                     has_thermal = 0;

    /* Initialize    */
    plist=cpl_propertylist_load(im_name, 0);

    /* Get the grating name */
    if ((sval = isaac_pfits_get_resolution(plist)) == NULL) {
        cpl_msg_error(cpl_func, "cannot get resolution from [%s]", im_name);
        cpl_propertylist_delete(plist);
        return -1;
    }
    strcpy(grat_name, sval);
    /* Get the central wavelength */
    wl_c = isaac_pfits_get_wlen(plist);
    if (cpl_error_get_code()) {
        cpl_msg_error(cpl_func, "cannot get central wlen from [%s]", im_name);
        cpl_propertylist_delete(plist);
        return -1;
    }
    wl_c *= 10000.0; /* microns to A */
    /* Get the filter used */
    if ((sval = isaac_pfits_get_filter(plist)) == NULL) {
        cpl_msg_error(cpl_func, "cannot get filter from [%s]", im_name);
        cpl_propertylist_delete(plist);
        return -1;
    }
    band = isaac_get_band(sval);
    cpl_propertylist_delete(plist);

    /* LW LR SK 2.20 Xe - Added Ar after testing
       - wl_c will currently only deviate from 2.22 in historical data  */
    if (grat_name[0] == 'L' && band == ISAAC_BAND_SK &&
        21900 <= wl_c) has_thermal = 1;

    /* LW LR SL 3.55 Xe+Ar */
    /* wl_c can in some (rare and actually unsupported) cases be lower */
    if (grat_name[0] == 'L' && band == ISAAC_BAND_SL &&
        34000 <= wl_c) has_thermal = 1;

    /* LW LR SH 3.55 Xe+Ar */
    /* This association is currently only relevant for historical data */
    /* The above limit for wl_c is chosen */
    if (grat_name[0] == 'L' && band == ISAAC_BAND_SH &&
        34000 <= wl_c && wl_c < 37000) has_thermal = 1;

    /* LW MR SK 2.35 - Added this after testing
       - 2.2 has no thermal background, while 2.26463 has */
    if (grat_name[0] == 'M' && band == ISAAC_BAND_SK &&
        22500 <= wl_c) has_thermal = 1;

    /* LW MR SL 3.30 */
    /* - and above (incl. 4.08) added after testing */
    if (grat_name[0] == 'M' && band == ISAAC_BAND_SL &&
        30000 <= wl_c) has_thermal = 1;

    /* LW MR SH Xe+Ar - Added after testing  */
    if (grat_name[0] == 'M' && band == ISAAC_BAND_SH &&
        32000 <= wl_c) has_thermal = 1;

    /* LW MR J+Block Xe+Ar - Added after testing  */
    if (grat_name[0] == 'M' && band == ISAAC_BAND_JBLOCK &&
        34000 <= wl_c) has_thermal = 1;

    cpl_msg_debug(cpl_func,
            "Has thermal: %d. Resol: %c. Lambda_c: %g. Filter: %s",
            has_thermal, grat_name[0], wl_c, isaac_get_filtername(band));

    return has_thermal;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Find out the slit width
  @param    filename    ISAAC fits file name 
  @return   the slit width in pixels or -1 in error case 
 */
/*----------------------------------------------------------------------------*/
double isaac_get_slitwidth(const char * filename)
{
    const char          *   sval;
    double                  slit_width;
    double                  pscale;
    cpl_propertylist    *   plist;

    /* Initialize    */
    plist=cpl_propertylist_load(filename, 0);

    /* Get the slit name used */
    if ((sval = isaac_pfits_get_opti1_id(plist)) == NULL) {
        cpl_msg_error(cpl_func, "cannot get slit used");
        cpl_propertylist_delete(plist);
        return -1.0;
    }

    /* Get the slit width in arcseconds */
    if (!strcmp(sval, "slit_1"))                 slit_width = 1;
    else if (!strcmp(sval, "slit_2"))            slit_width = 2;
    else if (!strcmp(sval, "slit_0.3_tilted"))   slit_width = 0.3;
    else if (!strcmp(sval, "slit_0.8"))          slit_width = 0.8;
    else if (!strcmp(sval, "slit_1.5"))          slit_width = 1.5;
    else if (!strcmp(sval, "slit_0.6_tilted"))   slit_width = 0.6;
    else {
        cpl_msg_error(cpl_func, "unrecognized slit");
        cpl_propertylist_delete(plist);
        return -1.0;
    }

    /* Get the pixelscale and convert arsec -> pixels */
    pscale = isaac_pfits_get_pixscale(plist);
    if (cpl_error_get_code()) {
        cpl_msg_error(cpl_func, "cannot get pixscale");
        cpl_propertylist_delete(plist);
        return -1.0;
    }
    cpl_propertylist_delete(plist);

    cpl_msg_debug(cpl_func, "Slit width = %g arcsec (%-.2f pixels)\n",
                  slit_width, slit_width / pscale);

    slit_width /= pscale;

    /* Return  */
    return slit_width;
}

/**@}*/


/*----------------------------------------------------------------------------*/
/**
  @brief    Count the positive-intensity lines in a given wavelength range
  @param    lines       catalog
  @param    wave_min    minimum wavelength
  @param    wave_max    maximum wavelength
  @return   the number of lines, or -1 on error
 */
/*----------------------------------------------------------------------------*/
static int isaac_wavelength_count_lines(const cpl_bivector * lines,
                                        double               wave_min,
                                        double               wave_max)
{
    const int      cat_nlines = cpl_bivector_get_size(lines);
    const double * px = cpl_bivector_get_x_data_const(lines);
    const double * py = cpl_bivector_get_y_data_const(lines);
    int            nb_lines = 0;
    int            i = 0;

    cpl_ensure(lines != NULL, CPL_ERROR_NULL_INPUT, -1);
    
    while (i < cat_nlines && px[i] < wave_min) i++;
    while (i < cat_nlines && px[i] < wave_max)
        if (py[i++] > 0) nb_lines++;

    return nb_lines;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Count the positive-intensity lines in a given wavelength range
  @param    lines       catalog
  @param    wave_min    minimum wavelength
  @param    wave_max    maximum wavelength
  @return   the number of lines, or -1 on error
 */
/*----------------------------------------------------------------------------*/
static int isaac_wavelength_count_linez(const cpl_bivector * lines,
                                        double               wave_min,
                                        double               wave_max) 
{
    const int      cat_nlines = cpl_bivector_get_size(lines);
    const double * px = cpl_bivector_get_x_data_const(lines);
    int            nb_lines = 0;
    int            i = 0;

    cpl_ensure(lines != NULL, CPL_ERROR_NULL_INPUT, -1);
    
    while (i < cat_nlines && px[i] < wave_min) i++;
    while (i < cat_nlines && px[i++] < wave_max) nb_lines++;

    return nb_lines;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Load the catalog
  @param    name        the lines names
  @param    oh          the oh catalog
  @param    ar          the ar catalog
  @param    xe          the xe catalog
  @return   1 newly allocated cpl_bivector or NULL in error case
 */
/*----------------------------------------------------------------------------*/
static cpl_bivector * isaac_wavelength_load_catalog(
        const char  *   name,
        const char  *   oh,
        const char  *   ar,
        const char  *   xe)
{
    cpl_bivector        *   sig;
    double              *   psigx;
    double              *   psigy;
    double              *   pcat_tab1_wave;
    double              *   pcat_tab1_emiss;
    int                     nlines;
    cpl_table           *   cat_tab1;
    cpl_table           *   cat_tab2;
    cpl_propertylist    *   reflist;
    int                     i;

    /* Check entries */
    if (name == NULL) return NULL;

    /* Handle the table names and load the tables */
    if (!strcmp(name, "oh")) {
        /* Check if the table is there */
        if (oh == NULL) {
            cpl_msg_error(cpl_func, "Please provide the OH lines catalog");
            return NULL;
        }
        /* Load the table */
        if ((cat_tab1 = cpl_table_load(oh, 1, 0)) == NULL) {
            cpl_msg_error(cpl_func, "Cannot load the catalog");
            return NULL;
        }
    } else if (!strcmp(name, "Xe")) {
        /* Check if the table is there */
        if (xe == NULL) {
            cpl_msg_error(cpl_func, "Please provide the XE lines catalog");
            return NULL;
        }
        /* Load the table */
        if ((cat_tab1 = cpl_table_load(xe, 1, 0)) == NULL) {
            cpl_msg_error(cpl_func, "Cannot load the XE catalog");
            return NULL;
        }
    } else if (!strcmp(name, "Ar")) {
        /* Check if the table is there */
        if (ar == NULL) {
            cpl_msg_error(cpl_func, "Please provide the AR lines catalog");
            return NULL;
        }
        /* Load the table */
        if ((cat_tab1 = cpl_table_load(ar, 1, 0)) == NULL) {
            cpl_msg_error(cpl_func, "Cannot load the AR catalog");
            return NULL;
        }
    } else if (!strcmp(name, "Xe+Ar")) {
        /* Check if the table is there */
        if (xe == NULL) {
            cpl_msg_error(cpl_func, "Please provide the XE lines catalog");
            return NULL;
        }
        /* Load the table */
        if ((cat_tab1 = cpl_table_load(xe, 1, 0)) == NULL) {
            cpl_msg_error(cpl_func, "Cannot load the XE catalog");
            return NULL;
        }
        /* Check if the table is there */
        if (ar == NULL) {
            cpl_msg_error(cpl_func, "Please provide the AR lines catalog");
            return NULL;
        }
        /* Load the table */
        if ((cat_tab2 = cpl_table_load(ar, 1, 0)) == NULL) {
            cpl_msg_error(cpl_func, "Cannot load the AR catalog");
            cpl_table_delete(cat_tab1);
            return NULL;
        }
        /* Merge the tables */
        if (cpl_table_insert(cat_tab1, cat_tab2,
                    cpl_table_get_nrow(cat_tab1)) != CPL_ERROR_NONE) {
            cpl_msg_error(cpl_func, "Cannot merge tables");
            cpl_table_delete(cat_tab1);
            cpl_table_delete(cat_tab2);
            return NULL;
        }
        cpl_table_delete(cat_tab2);
    } else return NULL;

    /* Sort the table */
    reflist = cpl_propertylist_new();
    cpl_propertylist_append_bool(reflist, ISAAC_COL_WAVELENGTH, 0);
    cpl_table_sort(cat_tab1, reflist);
    cpl_propertylist_delete(reflist);

    /* Create the bivector */
    nlines = cpl_table_get_nrow(cat_tab1);
    sig = cpl_bivector_new(nlines);
    psigx = cpl_bivector_get_x_data(sig);
    psigy = cpl_bivector_get_y_data(sig);
    pcat_tab1_wave = cpl_table_get_data_double(cat_tab1, ISAAC_COL_WAVELENGTH);
    pcat_tab1_emiss = cpl_table_get_data_double(cat_tab1, ISAAC_COL_EMISSION); 
    for (i=0; i<nlines; i++) {
        psigx[i] = pcat_tab1_wave[i];
        psigy[i] = pcat_tab1_emiss[i];
    }
    cpl_table_delete(cat_tab1);
    return sig;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Get the band 
  @param    filter  from the header
  @return   the band or ISAAC_BAND_UNKNOWN
 */
/*----------------------------------------------------------------------------*/
static isaac_band isaac_get_band(const char * filter)
{
    if (filter == NULL) return ISAAC_BAND_UNKNOWN;
    if (!strcmp(filter, "Z")) return ISAAC_BAND_Z;
    if (!strcmp(filter, "SZ")) return ISAAC_BAND_SZ;
    if (!strcmp(filter, "Js")) return ISAAC_BAND_JS;
    if (!strcmp(filter, "J")) return ISAAC_BAND_J;
    if (!strcmp(filter, "J+Block")) return ISAAC_BAND_JBLOCK;
    if (!strcmp(filter, "SH")) return ISAAC_BAND_SH;
    if (!strcmp(filter, "H")) return  ISAAC_BAND_H;
    if (!strcmp(filter, "Ks")) return ISAAC_BAND_KS;
    if (!strcmp(filter, "SK")) return ISAAC_BAND_SK;
    if (!strcmp(filter, "K")) return ISAAC_BAND_K;
    if (!strcmp(filter, "SL")) return ISAAC_BAND_SL;
    if (!strcmp(filter, "L")) return ISAAC_BAND_L;
    if (!strcmp(filter, "M")) return ISAAC_BAND_M;
    return ISAAC_BAND_UNKNOWN;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Get the filter name
  @param    filter  from the header
  @return   the band or ISAAC_BAND_UNKNOWN
 */
/*----------------------------------------------------------------------------*/
static const char * isaac_get_filtername(isaac_band band)
{
    if (band == ISAAC_BAND_UNKNOWN) return NULL;
    if (band == ISAAC_BAND_Z) return "Z";
    if (band == ISAAC_BAND_SZ) return "SZ";
    if (band == ISAAC_BAND_JS) return "Js";
    if (band == ISAAC_BAND_J) return "J";
    if (band == ISAAC_BAND_JBLOCK) return "J+Block";
    if (band == ISAAC_BAND_SH) return "SH";
    if (band == ISAAC_BAND_H) return "H";
    if (band == ISAAC_BAND_KS) return "Ks";
    if (band == ISAAC_BAND_SK) return "SK";
    if (band == ISAAC_BAND_K) return "K";
    if (band == ISAAC_BAND_SL) return "SL";
    if (band == ISAAC_BAND_L) return "L";
    if (band == ISAAC_BAND_M) return "M";
    
    return NULL;
}


/*----------------------------------------------------------------------------*/
/**
  @brief  EXPERIMENTAL: Distance from the pixel with the peak to the real peak
  @param  cat          The lines catalog
  @param  spec         The spectrum from the collapsed image
  @param  pix_low      First pixel to consider
  @param  ipix         the pixel with the (locally) maximum intensity
  @param  pix_high     Last pixel to consider
  @param  iline        The ind of the current emission line in the list
  @param  nline        The number of the current emission line
  @param  slit_width   Width (in pixels) of the slit used.
  @param  disprel      Dispersion 3rd degree polynomial
  @param  poffset      pointer to the subpixel offset
  @return -1 on error, otherwise the absolute value of the offset

*/
/*----------------------------------------------------------------------------*/
static double get_line_offset(
        const cpl_bivector      *   cat,
        const cpl_vector        *   spec,
        int                         pix_low,
        int                         ipix,
        double                      isub,
        int                         pix_high,
        int                         iline,
        int                         nline,
        double                      slit_width,
        const double            *   disprel,
        double                  *   poffset,
        double                  *   pmaxval)
{
    double              offset;
    const double    *   px = cpl_bivector_get_x_data_const(cat);
    const double    *   py = cpl_bivector_get_y_data_const(cat);
    const double        wl = px[iline];
    /* The expected location with sub-pixel precision */
    const double        spix = 1+ipix + isub;
    double              lastval;
    double              relint;
    const double        sigma  = slit_width * SLITWIDTH_TO_SIGMA;
    /* Tolerance for noise at emission-line center */
    const double        centernoise = 1.25;
    const int           mpix = pix_high-pix_low+1;
    const int           maxdist = slit_width;
    const double    *   pspec = cpl_vector_get_data_const(spec);
    double              maxval;
    int                 maxpos;
    int                 imin, imax;
    int                 i;

    cpl_ensure(pmaxval != NULL, CPL_ERROR_NULL_INPUT, -1);
    
    *pmaxval = -1;
    if (sigma <= 0) return -1;
    if (mpix <= 2) return -1;

    if (ipix >= cpl_vector_get_size(spec)) {
        cpl_msg_error(cpl_func, "IPIX(mpix=%d): %d <= ipix=%d <= %d <= %d",
                      mpix, pix_low, ipix, pix_high,
                      (int)cpl_vector_get_size(spec));
        return -1.0;
    }

    maxval = pspec[ipix];
    maxpos = ipix;
    for (i=pix_low; i<=pix_high; i++)
        if (pspec[i] > maxval) maxval = pspec[maxpos = i];
    *pmaxval = maxval;

    if (maxval == 0 ) {
        cpl_msg_debug(cpl_func, "0LINE %d (%g) at pixel: %d <= %g <= %d",
                nline, wl, 1+pix_low, spix, 1+pix_high);
        return -1;
    }
    lastval = maxval;
    imax    = maxpos;
    while (imax+1 <= pix_high && (pspec[imax+1] < lastval || pspec[imax+1] ==0
          || (imax-maxpos < maxdist && pspec[imax+1] < centernoise * lastval)))
        lastval = pspec[++imax]; 

    lastval = maxval;
    imin    = maxpos;

    while (imin-1 >= pix_low && (pspec[imin-1] < lastval || pspec[imin-1] == 0
         || (maxpos-imin < maxdist && pspec[imin-1] < centernoise * lastval)))
        lastval = pspec[--imin];

    offset = isaac_wavelength_centroid(spec, imin, imax);

    if (abs(maxpos - ipix) > maxdist ) {
        /* This must be another (unidenfitied) line which is
           brighter than the catalog line in this interval */
        const int plow  = maxpos > ipix ? pix_low  : (maxpos + ipix)/2; 
        const int phigh = maxpos < ipix ? pix_high : (maxpos + ipix)/2;
        if (offset < 0) {
            cpl_msg_debug(cpl_func, "LINE %d (%g) at pixel: %d <= %g/%d <= %d", 
                    nline, WAVELEN(disprel,1+maxpos), 1+pix_low, spix, 
                    1+maxpos, 1+pix_high);
        } else {
            offset -= isub;
            cpl_msg_debug(cpl_func, "LIne %d (%g) at pixel: %d <= %g/%d/%g <= %d",
                    nline, WAVELEN(disprel,1+offset+imin), 1+pix_low,
                    spix, 1+maxpos, 1+offset+imin, 1+pix_high);
        }
        /* Try to find the catalog line */
        return get_line_offset(cat, spec, plow, ipix, isub, phigh, iline, 
                               nline, slit_width, disprel, poffset, pmaxval); 
    }

    relint = 100*py[iline]/(sigma*sqrt(2*4*atan(1))*maxval);

    if (offset < 0) {
        /* maxpos too close to boundary or something like that */
        cpl_msg_debug(cpl_func, "LINe %d (%g) at pixel: %d/%d <= %g/%d <= %d/%d "
                  "(%4.2f%%)\n", nline, wl, 1+pix_low, 1+imin, spix, 1+maxpos,
                  1+imax, 1+pix_high, relint);
        return -1;
    }
    offset -= isub;
    *poffset = offset + imin - ipix;

    cpl_msg_debug(cpl_func,
            "Line %d (%g) at pixel: %d/%d <= %g/%d/%g <= %d/%d (%4.2f%%) %g",
            nline, wl, 1+pix_low, 1+imin, spix, 1+maxpos, 1+offset + imin,
            1+imax, 1+pix_high, relint, *poffset);

    return fabs(*poffset);
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Find out a vector centroid to subpixel precision.
  @param    vec     the vector
  @param    start   start ind to consider (0 -> nelem-1)
  @param    stop   stop ind to consider (0 -> nelem-1)
  @return   Centroid position as a double, or -1 on error.

  There must be at least 5 pixels on either side of the maximum pixelvalue.
  The position of the peak is located to subpixel precision by
  simply weighting positions with pixelvalues.
 */
/*----------------------------------------------------------------------------*/
static double isaac_wavelength_centroid(
        const cpl_vector    *   vec,
        int                     start,
        int                     stop)
{
    const double    *   pvec;
    int                 half_centroid_domain = 5;
    double              min = 0;
    double              max;
    double              centroid;
    double              weights;
    int                 i, maxpos;

    if (vec==NULL) return -1.0;
    pvec = cpl_vector_get_data_const(vec);

    /* Search for the maximum pixel value on the line */
    max = pvec[start];
    maxpos = start;
    for (i=start; i<=stop; i++) {
        if (pvec[i]>max) {
            max = pvec[i];
            maxpos = i;
        }
    }

    if (maxpos < start + half_centroid_domain ||
             maxpos >= stop + 1 - half_centroid_domain) return -1.0;

    /* Centroiding is only defined for non-negative intensities. If the
       centroiding region has negative intensities then find the minimum
       and offset the signal by this minimum */
    for (i=maxpos-half_centroid_domain; i<=maxpos+half_centroid_domain; i++)
        if (pvec[i] < min) min = pvec[i];

    /* The centroid pos is the weighted average around the max pixel */
    centroid = 0.0;
    weights  = 0.0;
    if (min < 0) {
        for (i=maxpos-half_centroid_domain;i<=maxpos+half_centroid_domain;i++) {
            centroid += (double)(pvec[i]-min) * (double)i;
            weights  += (double)(pvec[i]-min);
        }
    } else {
        for (i=maxpos-half_centroid_domain;i<=maxpos+half_centroid_domain;i++) {
            centroid += (double)(pvec[i]) * (double)i;
            weights  += (double)(pvec[i]);
        }
    }

    if (fabs(weights)<fabs(centroid)*1e-5 ) centroid = -1.0;
    else centroid /= weights;

    return centroid - start;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Locate the emissionlines with sub-pixel precision
  @param  cat           The lines catalog
  @param  spec          The spectrum from the collapsed image
  @param  discard_le    Number of pixels to discard on the left
  @param  discard_ri    Number of pixels to discard on the left
  @param  npix          Number of pixels in spectrum
  @param  slit_width    Width (in pixels) of the slit used.
  @param  disprel       Dispersion 3rd degree polynomial
  @return void
*/
/*----------------------------------------------------------------------------*/
static computed_disprel * spectro_refine_solution(
        const cpl_bivector  *   cat,
        const cpl_vector    *   spec,
        int                     discard_le,
        int                     discard_ri,
        int                     npix,
        double                  slit_width,
        const double        *   disprel)
{
    computed_disprel    *   solution;
    const int               istart = discard_le > 0 ? discard_le : 0;
    const int               istop  = discard_ri > 0 ? npix-discard_ri : npix;

    const double            wl_min  = WAVELEN(disprel, 0 + 0.5);
    const double            wl_max  = WAVELEN(disprel, npix  + 1.5);

    const int               nlines = cpl_bivector_get_size(cat);
    const double          * px = cpl_bivector_get_x_data_const(cat);
    const double          * py = cpl_bivector_get_y_data_const(cat);

    double                  isub = 0;
    double                  sum_offset = 0;
    double                  sum_aboffs = 0;
    double                  sum_sqoffs = 0;
    double                  offset;
    double                  maxval;
    double                  maxint = 0;
    double                  faint = 0;

    int                     pix_high = 0;
    int                     ical = 0;
    int                     iline = 0;

    const int emil  = isaac_wavelength_count_lines(cat, wl_min, wl_max);
    const int emilz = isaac_wavelength_count_linez(cat, wl_min, wl_max);
    int                     nline  = 0;
    int                     nfound = 0;
    int                     nzero  = 0;

   
    if (emil < 1) return NULL;

    solution = cpl_malloc(sizeof(*solution));

    /* Find first line within range */
    while (iline < nlines && px[iline] < wl_min) iline++;

    /* Do not try to locate the very faint lines on a very crowded spectrum */
    if (emilz * (11 + slit_width) > npix) {
        int i;

        /* Find the maximum intensity */
        for (i=iline; i<nlines; i++) {
            if (px[i] > wl_max) break;
            if (py[i] > maxint) maxint = py[i];
        }
        /* Ignore lines with intensity less than 0.01 of the maximum */
        faint = 0.01;
        cpl_msg_debug(cpl_func,
                "No detection of faint lines (I < %g) in crowded spectrum: %g",
                maxint * faint, npix/(emilz * slit_width));
    }

    while (pix_high < npix - 1) {
        double          wl;
        const double    isub_prev  = isub;
        const int       icalprev   = ical;
        const int       pix_low    = pix_high;

        /* Do not try to locate the very faint lines
           - zero-intensity means unknown intensity - not to be ignored
           - also ignore the duplicate OH sky lines */
        while (iline<nlines && 0 < py[iline] &&
               (py[iline] < maxint * faint ||
                (py[iline] == py[iline-1] &&
                (px[iline] - px[iline-1])
                < 0.5*WAVEDLT(disprel, 1+icalprev)*slit_width))) {
            cpl_msg_debug(cpl_func, "Skipping line(%d): %g %g", iline,
                    px[iline], py[iline]);
            iline++;
        }

        if (iline < nlines &&
           /* The wavelength of the emission line */
           (wl = px[iline]) < wl_max) {
            double wl_low, wl_high;
            double delta;
            int i=3;

            /* Find the expected pixel of the emission line
                - sample nr. x (with ind i = x-1) has wavelengths
                from p(x-0.5) to p(x+0.5) */

            /* More than one wavelength can belong to a pixel */
            ical--;  

            while (++ical < npix-1 && WAVELEN(disprel, ical+1 + 0.5) < wl);
            wl_low  = WAVELEN(disprel, ical+1 - 0.5);
            wl_high = WAVELEN(disprel, ical+1 + 0.5); 

            /* -0.5 <= isub < 0.5 */
            isub = (wl - wl_low) / (wl_high - wl_low) - 0.5;
            /* Just since its easy: Determine isub with maximum precision
               - using Newton-Raphson */
            do {
                delta = (WAVELEN(disprel, ical+1+isub)-wl)
                      /  WAVEDIF(disprel, ical+1+isub); 
                isub -= delta;
            } while (--i && fabs(delta) > DBL_EPSILON);
            /* At this point delta is most likely true 0 :-) */
            pix_high = (ical + icalprev)/2;
        } else {
            pix_high = npix - 1;
        }

        if (nline > 0) {
            if (get_line_offset(cat, spec, pix_low, icalprev, isub_prev,
                                pix_high, iline-1, nline, slit_width,
                                disprel, &offset, &maxval) >=0) {
                if (istart <= icalprev && icalprev <= istop) {
                    /* Only include lines that are not in the discard range */
                    sum_offset += offset;
                    sum_aboffs += fabs(offset);
                    sum_sqoffs += offset * offset;
                    nfound++;
                }
            }
            if (maxval == 0) nzero++;
        }
        iline++;
        nline++;
    }
    solution->dlines = nfound;
    solution->clines = emilz;
    solution->rms    = -1;
    solution->mean   = -1;

    if (nfound > 0) {
        double stdev = nfound ==1 ? 0 :
                       (sum_sqoffs-sum_offset*sum_offset/nfound)/(nfound-1);
        stdev = stdev > 0 ? sqrt(stdev) : 0;
        
        cpl_msg_debug(cpl_func,
                "Mean & RMS pixel-offset on calibration (%d:%d:%d:%d): %g %g", 
                emilz, emilz-emil,nzero, nfound, sum_aboffs/nfound, stdev);

        /* The bias-corrected standard deviation */
        solution->rms    = stdev;
        /* The mean deviation */
        solution->mean   = sum_aboffs/nfound;
    }
    return solution;
}
