/* $Id: isaac_spc_startrace.c,v 1.51 2013-03-12 08:06:48 llundin Exp $
 *
 * This file is part of the ISAAC Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA
 */

/*
 * $Author: llundin $
 * $Date: 2013-03-12 08:06:48 $
 * $Revision: 1.51 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <math.h>
#include <float.h>
#include <cpl.h>

#include "irplib_plugin.h"
#include "irplib_utils.h"
#include "irplib_spectrum.h"

#include "isaac_utils.h"
#include "isaac_pfits.h"
#include "isaac_dfs.h"

/*-----------------------------------------------------------------------------
                                Define
 -----------------------------------------------------------------------------*/

#define RECIPE_STRING "isaac_spc_startrace"

#define Z_LR_LEFT_REJ       300
#define Z_LR_RIGHT_REJ      325
#define SZ_LR_LEFT_REJ      300
#define SZ_LR_RIGHT_REJ     325
#define J_LR_LEFT_REJ       200
#define J_LR_RIGHT_REJ      200
#define SH_LR_LEFT_REJ      150 
#define SH_LR_RIGHT_REJ     175
#define SK_LR_LEFT_REJ      150
#define SK_LR_RIGHT_REJ     175
#define MR_LEFT_REJ         30
#define MR_RIGHT_REJ        30

/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

static cpl_error_code isaac_spc_startrace_reduce(cpl_frameset *,
                                                 const cpl_frameset *,
                                                 const cpl_matrix *, char, 
                                                 const cpl_parameterlist *);

static cpl_size * isaac_spc_startrace_resol(cpl_frameset *);
static cpl_imagelist * isaac_spc_startrace_load(const cpl_frameset *);
static cpl_matrix * isaac_spc_startrace_starpos(const cpl_frameset *);
static cpl_polynomial * isaac_spc_startrace_shape(cpl_image *, double, char, 
        double *);
static cpl_polynomial * isaac_spc_startrace_distor(cpl_polynomial **, 
        cpl_vector *, int, int);
static cpl_error_code isaac_spc_startrace_save(cpl_frameset *,
                                               const cpl_table *,
                                               const cpl_matrix *,
                                               const cpl_vector *,
                                               cpl_polynomial **,
                                               const cpl_frameset *,
                                               char, 
                                               const cpl_parameterlist *);
 

cpl_recipe_define(isaac_spc_startrace, ISAAC_BINARY_VERSION,
                  "Lars Lundin", PACKAGE_BUGREPORT, "2002, 2003, 2008", 
                  "ISAAC Spectro startrace recipe",
                  RECIPE_STRING " -- ISAAC Spectro startrace recipe\n"
                  "The files listed in the Set Of Frames (sof-file) " 
                  "must be tagged:\n"
                  "raw-file.fits "ISAAC_SPC_STARTRACE_IM_RAW" or\n"
                  "raw-file.fits "ISAAC_SPC_STARTRACE_SP_RAW"\n"
                  "Given N imaging frames, the spectro frames must consist of "
                  "N (or zero) low resolution frames and "
                  "N (or zero) medium resolution frames.");

/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

static struct {
    /* Inputs */
    cpl_size    degree;
    int         spec_width;
    int         reject_left;
    int         reject_right;
    int         display;
    /* Outputs */
    int         arm;
    isaac_band  band;
    double      corr_is1;
    double      corr_is2;
    double      corr_is3;
    double      shapes_fit_q;
    double      dist1;
    double      distx;
    double      disty;
    double      distxy;
    double      distxx;
    double      distyy;
} isaac_spc_startrace_config;

/*-----------------------------------------------------------------------------
                                Functions code
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    Fill the recipe parameterlist
  @param    self  The parameterlist
  @return   0 iff everything is ok

 */
/*----------------------------------------------------------------------------*/
static
cpl_error_code isaac_spc_startrace_fill_parameterlist(cpl_parameterlist * self)
{
    const char * context = PACKAGE "." RECIPE_STRING;
    cpl_error_code err;

    cpl_ensure_code(self, CPL_ERROR_NULL_INPUT);

    /* Fill the parameters list */

    /* --degree */
    err = irplib_parameterlist_set_int(self, PACKAGE, RECIPE_STRING,
                                       "degree", 3, "deg",
                                       context, "polynomial degree");
    cpl_ensure_code(!err, err);

    /* --spec_width */
    err = irplib_parameterlist_set_int(self, PACKAGE, RECIPE_STRING,
                                       "spectrum_width", 40, "spec_width",
                                       context, "spectrum width");
    cpl_ensure_code(!err, err);

     /* --reject_left */
    err = irplib_parameterlist_set_int(self, PACKAGE, RECIPE_STRING,
                                       "reject_left", -1, NULL,
                                       context, "left part rejection");
    cpl_ensure_code(!err, err);

     /* --reject_right */
    err = irplib_parameterlist_set_int(self, PACKAGE, RECIPE_STRING,
                                       "reject_right", -1, NULL,
                                       context, "right part rejection");
    cpl_ensure_code(!err, err);

    /* --display */
    err = irplib_parameterlist_set_bool(self, PACKAGE, RECIPE_STRING,
                                        "display", CPL_FALSE, NULL, context,
                                        "Flag to make plots");
    cpl_ensure_code(!err, err);

    return CPL_ERROR_NONE;
}


/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    The recipe data reduction part is implemented here 
  @param    parlist     the parameters list
  @param    framelist   the frames list
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int isaac_spc_startrace(cpl_frameset            * framelist,
                               const cpl_parameterlist * parlist)
{
    const char          *   sval;
    cpl_size            *   labels = NULL;
    const cpl_frame     *   cur_frame;
    cpl_propertylist    *   plist = NULL;
    cpl_frameset        *   imframes = NULL;
    cpl_frameset        *   spframes = NULL;
    cpl_frameset        *   spframes_lr = NULL;
    cpl_frameset        *   spframes_mr = NULL;
    cpl_matrix          *   star_pos = NULL;
    int                     ndone = 0;
    

    bug_if(0);

    /* Retrieve input parameters */

    /* Polynomial degree */
    isaac_spc_startrace_config.degree
        = irplib_parameterlist_get_int(parlist, PACKAGE, RECIPE_STRING,
                                       "degree");
    /* Spectrum width */
    isaac_spc_startrace_config.spec_width
        = irplib_parameterlist_get_int(parlist, PACKAGE, RECIPE_STRING,
                                       "spectrum_width");
    /* Rejection parameters */
    isaac_spc_startrace_config.reject_left
        = irplib_parameterlist_get_int(parlist, PACKAGE, RECIPE_STRING,
                                       "reject_left");
    isaac_spc_startrace_config.reject_right
        = irplib_parameterlist_get_int(parlist, PACKAGE, RECIPE_STRING,
                                       "reject_right");
    /* Display */
    isaac_spc_startrace_config.display
        = irplib_parameterlist_get_bool(parlist, PACKAGE, RECIPE_STRING,
                                        "display");

    /* Identify the RAW and CALIB frames in the input frameset */
    skip_if (isaac_dfs_set_groups(framelist));
 
    /* Get the arm used (SW or LW) */
    cur_frame = cpl_frameset_get_position_const(framelist,0);
    error_if(cur_frame == NULL, CPL_ERROR_DATA_NOT_FOUND,
             "Could not get 1st frame from frameset");
    plist = cpl_propertylist_load(cpl_frame_get_filename(cur_frame), 0);
    any_if("Could not load propertylist from 1st frame");

    sval = isaac_pfits_get_arm(plist);
    any_if("Could not get the arm");

    if (sval[0] == 'S') {
        isaac_spc_startrace_config.arm = 1;
    } else if (sval[0] == 'L') {
        isaac_spc_startrace_config.arm = 2;
    } else {
        error_if(1, CPL_ERROR_UNSUPPORTED_MODE, "Unsupported arm: %s", sval);
    }

    /* Get the band */
    sval = isaac_pfits_get_filter(plist);
    any_if("Could not get the filter");

    isaac_spc_startrace_config.band = isaac_get_bbfilter(sval);
    any_if("Could not get the band");

    cpl_propertylist_empty(plist);
    
    /* Retrieve raw frames */
    imframes = isaac_extract_frameset(framelist, ISAAC_SPC_STARTRACE_IM_RAW);
    error_if(imframes == NULL, CPL_ERROR_DATA_NOT_FOUND, "Could not find the "
             "imframes in the input list");

    spframes = isaac_extract_frameset(framelist, ISAAC_SPC_STARTRACE_SP_RAW);
    error_if(spframes == NULL, CPL_ERROR_DATA_NOT_FOUND, "Could not find the "
             "spframes in the input list");

    /* Separate LR and MR spectra */
    labels = isaac_spc_startrace_resol(spframes);
    error_if(labels == NULL, cpl_error_get_code(), "Could not differentiate "
             "LR and MR");

    spframes_lr = cpl_frameset_extract(spframes, labels, 1);
    spframes_mr = cpl_frameset_extract(spframes, labels, 2);
    cpl_free(labels);
    labels = NULL;
    
    /* Get the star positions */
    cpl_msg_info(cpl_func, "Get the star positions");
    star_pos = isaac_spc_startrace_starpos(imframes);
    error_if(star_pos == NULL, cpl_error_get_code(), "Could not find star the "
             "positions");

    cpl_frameset_delete(imframes);
    imframes = NULL;

    error_if (spframes_lr == NULL && spframes_mr == NULL,
              CPL_ERROR_UNSUPPORTED_MODE, "The %d spectro frames are not LR "
              "nor MR", (int)cpl_frameset_get_size(spframes));
    cpl_frameset_delete(spframes);
    spframes = NULL;

    if (spframes_lr != NULL) {
        cpl_errorstate prestate = cpl_errorstate_get();
        /* Reduction in LR */
        cpl_msg_info(cpl_func, "Reducing the LR spectra");
        if (isaac_spc_startrace_reduce(framelist, spframes_lr, star_pos,
                                       'L', parlist)) {
            irplib_error_recover(prestate, "Could not reduce the LR spectra");
        } else {
            ndone++;
        }
        cpl_frameset_delete(spframes_lr);
        spframes_lr = NULL;
    }
    if (spframes_mr != NULL) {
        cpl_errorstate prestate = cpl_errorstate_get();
        /* Reduction in MR */
        cpl_msg_info(cpl_func, "Reducing the MR spectra");
        if (isaac_spc_startrace_reduce(framelist, spframes_mr, star_pos,
                                       'M', parlist)) {
            irplib_error_recover(prestate, "Could not reduce the MR spectra");
        } else {
            ndone++;
        }
        cpl_frameset_delete(spframes_mr);
        spframes_mr = NULL;
    }

    error_if(!ndone, CPL_ERROR_ILLEGAL_INPUT, "Could not reduce the spectra");

    end_skip;

    cpl_free(labels);
    cpl_propertylist_delete(plist);
    cpl_frameset_delete(imframes);
    cpl_frameset_delete(spframes);
    cpl_frameset_delete(spframes_lr);
    cpl_frameset_delete(spframes_mr);
    cpl_matrix_delete(star_pos);

    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    The spectra reduction is applied here
  @param    framelist   the input frame set
  @param    spframes    the spectrum frames
  @param    star_pos    the star positions
  @param    resol       the resolution ('L' or 'M')
  @param    parlist     the input parameter list
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static
cpl_error_code isaac_spc_startrace_reduce(cpl_frameset            * framelist,
                                          const cpl_frameset      * spframes,
                                          const cpl_matrix        * star_pos,
                                          char                      resol,
                                          const cpl_parameterlist * parlist)
{
    /* Load the input images */
    cpl_imagelist   * ilist = isaac_spc_startrace_load(spframes);
    const int         nframes = cpl_frameset_get_size(spframes);
    cpl_vector      * spec_pos = cpl_vector_new(nframes);
    double          * pspec_pos = cpl_vector_get_data(spec_pos);
    cpl_polynomial  * corresp = cpl_polynomial_new(1);
    cpl_polynomial ** shapes = cpl_calloc(nframes, sizeof(cpl_polynomial*));
    cpl_vector      * fit_q = cpl_vector_new(nframes);
    double          * pfit_q = cpl_vector_get_data(fit_q);
    const int         nx = cpl_image_get_size_x(cpl_imagelist_get(ilist, 0));
    cpl_polynomial  * distor = NULL;
    cpl_table       * out_tab = NULL;
    cpl_size          power[2];
    const cpl_size    maxdeg = 2;
    cpl_size          i;

    bug_if(0);
    bug_if(framelist == NULL);
    bug_if(star_pos == NULL);
    bug_if(parlist == NULL);

    error_if(cpl_matrix_get_ncol(star_pos) != nframes,
             CPL_ERROR_INCOMPATIBLE_INPUT, "Number of spectro frames (%d) "
             "differ from number of star positions (%d)", nframes,
             (int)cpl_matrix_get_ncol(star_pos));

    /* Find the spectra positions */
    for (i=0; i<nframes; i++) {
        const cpl_image * image = cpl_imagelist_get_const(ilist, i);
        const double mean = cpl_image_get_mean(image);
        const double min_bright = 20.0 * mean;

        error_if (irplib_spectrum_find_brightest(image, 0, NO_SHADOW,
                                                 min_bright, 0, &(pspec_pos[i])),
                  cpl_error_get_code(), "Could not detect spectrum in frame %d "
                  "of %d with mean flux=%g", (int)i+1, nframes, mean);
    }

    /* Compute spectra / star correspondance */
    error_if (cpl_polynomial_fit(corresp, star_pos, NULL, spec_pos, NULL,
                                 CPL_FALSE, NULL, &maxdeg), cpl_error_get_code(),
              "Could not fit a %d degree polynomial to the %d positions",
              (int)maxdeg, nframes);

    /* Store the result coefficients */
    i = 0;
    isaac_spc_startrace_config.corr_is1 = cpl_polynomial_get_coeff(corresp, &i);
    i = 1;
    isaac_spc_startrace_config.corr_is2 = cpl_polynomial_get_coeff(corresp, &i);
    i = 2;
    isaac_spc_startrace_config.corr_is3 = cpl_polynomial_get_coeff(corresp, &i);
    cpl_polynomial_delete(corresp);
    corresp = NULL;
    
    /* Compute the spectra shapes */
    for (i=0; i<nframes; i++) {
        shapes[i] = isaac_spc_startrace_shape(cpl_imagelist_get(ilist, i),
                                              cpl_vector_get(spec_pos, i), 
                                              resol, 
                                              &(pfit_q[i]));

        error_if (shapes[i] == NULL, cpl_error_get_code(), "Could not get the "
                  "shape of frame %d of %d", (int)i+1, nframes);

        cpl_msg_info(cpl_func, "Spectrum %d : Y = %g; Fit mse = %g", (int)i+1,
                     cpl_vector_get(spec_pos, i), pfit_q[i]);
    }
    isaac_spc_startrace_config.shapes_fit_q = cpl_vector_get_median_const(fit_q);
    cpl_vector_delete(fit_q);
    fit_q = NULL;
    cpl_imagelist_delete(ilist);
    ilist = NULL;

    /* Compute the 2d distortion */
    distor = isaac_spc_startrace_distor(shapes, spec_pos, nframes, nx);
    error_if (distor == NULL, cpl_error_get_code(),  "Could not compute the "
              "distortion");
   
    /* Create the table */
    out_tab = cpl_table_new(6);
    cpl_table_new_column(out_tab, "Degree_of_x", CPL_TYPE_INT);
    cpl_table_new_column(out_tab, "Degree_of_y", CPL_TYPE_INT);
    cpl_table_new_column(out_tab, "poly2d_coef", CPL_TYPE_DOUBLE);
    power[0] = 0; power[1] = 0;
    cpl_table_set_int(out_tab, "Degree_of_x", 0, power[0]);
    cpl_table_set_int(out_tab, "Degree_of_y", 0, power[1]);
    isaac_spc_startrace_config.dist1 = cpl_polynomial_get_coeff(distor, power);
    cpl_table_set_double(out_tab, "poly2d_coef", 0,
            isaac_spc_startrace_config.dist1);
    power[0] = 1; power[1] = 0;
    cpl_table_set_int(out_tab, "Degree_of_x", 1, power[0]);
    cpl_table_set_int(out_tab, "Degree_of_y", 1, power[1]);
    isaac_spc_startrace_config.distx = cpl_polynomial_get_coeff(distor, power);
    cpl_table_set_double(out_tab, "poly2d_coef", 1,
            isaac_spc_startrace_config.distx);
    power[0] = 0; power[1] = 1;
    cpl_table_set_int(out_tab, "Degree_of_x", 2, power[0]);
    cpl_table_set_int(out_tab, "Degree_of_y", 2, power[1]);
    isaac_spc_startrace_config.disty = cpl_polynomial_get_coeff(distor, power);
    cpl_table_set_double(out_tab, "poly2d_coef", 2,
            isaac_spc_startrace_config.disty);
    power[0] = 1; power[1] = 1;
    cpl_table_set_int(out_tab, "Degree_of_x", 3, power[0]);
    cpl_table_set_int(out_tab, "Degree_of_y", 3, power[1]);
    isaac_spc_startrace_config.distxy = cpl_polynomial_get_coeff(distor, power);
    cpl_table_set_double(out_tab, "poly2d_coef", 3,
            isaac_spc_startrace_config.distxy);
    power[0] = 2; power[1] = 0;
    cpl_table_set_int(out_tab, "Degree_of_x", 4, power[0]);
    cpl_table_set_int(out_tab, "Degree_of_y", 4, power[1]);
    isaac_spc_startrace_config.distxx = cpl_polynomial_get_coeff(distor, power);
    cpl_table_set_double(out_tab, "poly2d_coef", 4,
            isaac_spc_startrace_config.distxx);
    power[0] = 0; power[1] = 2;
    cpl_table_set_int(out_tab, "Degree_of_x", 5, power[0]);
    cpl_table_set_int(out_tab, "Degree_of_y", 5, power[1]);
    isaac_spc_startrace_config.distyy = cpl_polynomial_get_coeff(distor, power);
    cpl_table_set_double(out_tab, "poly2d_coef", 5,
            isaac_spc_startrace_config.distyy);

    cpl_polynomial_delete(distor);
    distor = NULL;

    /* Display result */
    cpl_msg_info(cpl_func, 
            "Distortion : Y(x,y) = a + bx + cy + dxy + exx + fyy");
    cpl_msg_info(cpl_func, "a = %g", isaac_spc_startrace_config.dist1);
    cpl_msg_info(cpl_func, "b = %g", isaac_spc_startrace_config.distx);
    cpl_msg_info(cpl_func, "c = %g", isaac_spc_startrace_config.disty);
    cpl_msg_info(cpl_func, "d = %g", isaac_spc_startrace_config.distxy);
    cpl_msg_info(cpl_func, "e = %g", isaac_spc_startrace_config.distxx);
    cpl_msg_info(cpl_func, "f = %g", isaac_spc_startrace_config.distyy);

    /* Write the product */
    skip_if (isaac_spc_startrace_save(framelist, out_tab, star_pos, spec_pos,
                                      shapes, spframes, resol, parlist));

    end_skip;

    cpl_polynomial_delete(distor);
    cpl_vector_delete(fit_q);
    cpl_polynomial_delete(corresp);
    if (shapes != NULL) {
        for (i = 0; i < nframes; i++) cpl_polynomial_delete(shapes[i]);
        cpl_free(shapes);
    }
    cpl_vector_delete(spec_pos);
    cpl_imagelist_delete(ilist);
    cpl_table_delete(out_tab);

    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    Get the reolutions of the spectral frames
  @param    spframes    the spectral frames
  @return   an int array with 1 for LR, 2 for MR, 0 for others, or NULL
  in error case
 */
/*----------------------------------------------------------------------------*/
static cpl_size * isaac_spc_startrace_resol(cpl_frameset * spframes)
{
    int                     nframes;
    cpl_size            *   labels;
    cpl_frame           *   cur_frame;
    cpl_propertylist    *   plist;
    const char          *   sval;
    int                     i;

    /* Test entries */
    if (spframes == NULL) return NULL;

    /* Initialise */
    nframes = cpl_frameset_get_size(spframes);

    /* Create output array */
    labels = (cpl_size*)cpl_malloc(nframes * sizeof(*labels));

    /* Loop on the frames */
    for (i=0; i<nframes; i++) {
        cur_frame = cpl_frameset_get_position(spframes, i);
        plist=cpl_propertylist_load(cpl_frame_get_filename(cur_frame), 0);
        /* Get the resolution */
        if ((sval = isaac_pfits_get_resolution(plist)) == NULL) {
            cpl_free(labels);
            cpl_msg_error(cpl_func,"cannot get the resolution");
            return NULL;
        }
        if (sval[0]=='L') labels[i] = 1;
        else if (sval[0]=='M') labels[i] = 2;
        else labels[i] = 0;
        cpl_propertylist_delete(plist);
    }
    return labels;
}
   
/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    get the positions of the stars
  @param    imframes    imaging frames
  @return   a bivector with the star positions or null in error case.
 */
/*----------------------------------------------------------------------------*/
static cpl_matrix * isaac_spc_startrace_starpos(const cpl_frameset * imframes)
{
    int                     nframes;
    cpl_matrix          *   ypositions;
    cpl_apertures       *   objects;
    double              *   ypos;
    cpl_imagelist       *   ilist;
    int                     i;

    /* Test entries */
    if (imframes == NULL) return NULL;

    /* Load the input images */
    if ((ilist = isaac_spc_startrace_load(imframes)) == NULL) return NULL;

    /* Initialise */
    nframes = cpl_frameset_get_size(imframes);

    /* Create bivector */
    ypositions = cpl_matrix_new(1, nframes);
    ypos = cpl_matrix_get_data(ypositions);

    /* Loop on the frames */
    for (i=0; i<nframes; i++) {
        double xpos;
        if ((objects = cpl_apertures_extract_sigma(cpl_imagelist_get(ilist, i),
                        2.0)) == NULL) {
            cpl_msg_error(cpl_func, "cannot detect any object");
            cpl_matrix_delete(ypositions);
            cpl_imagelist_delete(ilist);
            return NULL;
        }
        cpl_apertures_sort_by_flux(objects);
        xpos = cpl_apertures_get_centroid_x(objects, 1);
        ypos[i] = cpl_apertures_get_centroid_y(objects, 1);
        cpl_apertures_delete(objects);
        cpl_msg_info(cpl_func, "Image %d: Position (%g, %g)", i+1,xpos, ypos[i]);
    }
    cpl_imagelist_delete(ilist);
    return ypositions;
}
 
/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    Load the imag list
  @param    inframes    frames
  @return   the image list or NULL in error case
 */
/*----------------------------------------------------------------------------*/
static cpl_imagelist * isaac_spc_startrace_load(const cpl_frameset * inframes)
{
    cpl_imagelist       *   ilist;
    int                     nframes;
    cpl_image           *   tmp_im;
    int                     i;

    /* Test entries */
    if (inframes == NULL) return NULL;

    /* Initialise */
    nframes = cpl_frameset_get_size(inframes);

    /* Load the images */
    if ((ilist = cpl_imagelist_load_frameset(inframes, CPL_TYPE_FLOAT, 1, 
                    0)) == NULL) return NULL;

    /* In LW that's it */
    if (isaac_spc_startrace_config.arm == 2) return ilist; 

    /* In SW, the frames are subtracted to remove the background */
    tmp_im = cpl_image_duplicate(cpl_imagelist_get(ilist, 0));
    for (i=0; i<nframes-1; i++) {
        if (cpl_image_subtract(cpl_imagelist_get(ilist, i),
                cpl_imagelist_get(ilist, i+1)) != CPL_ERROR_NONE) {
            cpl_msg_error(cpl_func, "cannot subtract images");
            cpl_imagelist_delete(ilist);
            cpl_image_delete(tmp_im);
            return NULL;
        }
    }
    /* Handle last frame */
    if (cpl_image_subtract(cpl_imagelist_get(ilist, nframes-1), 
            tmp_im) != CPL_ERROR_NONE) {
        cpl_msg_error(cpl_func, "cannot subtract images");
        cpl_imagelist_delete(ilist);
        cpl_image_delete(tmp_im);
        return NULL;
    }
    cpl_image_delete(tmp_im);

    /* Set negative pixels to 0  */
    cpl_imagelist_threshold(ilist, 0.0, DBL_MAX, 0.0, 0.0);
    
    return ilist;
}

/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    Get the shape of a spectrum
  @param    in      an image
  @param    ypos    the y position of the spectrum
  @param    resol   the resolution ('L' or 'M') 
  @param    fit_quality     the fit quality
  @return   the polynomial with the spectrum shape or NULL
 */
/*----------------------------------------------------------------------------*/
static cpl_polynomial * isaac_spc_startrace_shape(
        cpl_image       *   in,
        double              ypos,
        char                resol,
        double          *   fit_quality)
{
    int                     reject_left, reject_right;
    cpl_image           *   filtered;
    int                     low_side, up_side;
    cpl_mask            *   kernel;
    cpl_matrix          *   fit_x;
    cpl_vector          *   fit_y;
    double              *   pfit_x;
    double              *   pfit_y;
    cpl_polynomial      *   fitted;
    cpl_bivector        *   toplot;
    int                     i;
    int                     nsamples;
    const cpl_boolean       sampsym = CPL_TRUE; /* fit_x is symmetric */

    /* Check entries */
    if ((in == NULL) || (ypos < 0) || (fit_quality == NULL)) return NULL;
   
    /* Initialise */
    reject_left = isaac_spc_startrace_config.reject_left;
    reject_right = isaac_spc_startrace_config.reject_right;
    low_side = (int)(ypos - (isaac_spc_startrace_config.spec_width)/2);
    up_side = low_side + isaac_spc_startrace_config.spec_width;

    /* Test spectrum position */
    if ((low_side < 1) || (up_side > cpl_image_get_size_y(in))) {
        cpl_msg_error(cpl_func, "spectrum too close to the image border");
        return NULL;
    }
   
    /* Get rejection parameters */
    if (isaac_spc_startrace_config.arm == 1) {
        /* SW */
        if (reject_left < 0) {
            if (resol == 'M') reject_left = MR_LEFT_REJ;
            else if (resol == 'L') {
                switch (isaac_spc_startrace_config.band) {
                    case ISAAC_BAND_Z:    reject_left = Z_LR_LEFT_REJ; break;
                    case ISAAC_BAND_SZ:   reject_left = SZ_LR_LEFT_REJ; break;
                    case ISAAC_BAND_JBLOCK:
                    case ISAAC_BAND_J:    reject_left = J_LR_LEFT_REJ; break;
                    case ISAAC_BAND_SH:   reject_left = SH_LR_LEFT_REJ; break;
                    case ISAAC_BAND_SK:   reject_left = SK_LR_LEFT_REJ; break;
                    default:
                        cpl_msg_warning(cpl_func, "unsupported filter");
                        reject_left = 0;
                        break;
                }
            }
        }
        if (reject_right < 0) {
            if (resol == 'M') reject_right = MR_RIGHT_REJ;
            else if (resol == 'L') {
                switch (isaac_spc_startrace_config.band) {
                    case ISAAC_BAND_Z:    reject_right = Z_LR_RIGHT_REJ; break;
                    case ISAAC_BAND_SZ:   reject_right = SZ_LR_RIGHT_REJ;break;
                    case ISAAC_BAND_JBLOCK:
                    case ISAAC_BAND_J:    reject_right = J_LR_RIGHT_REJ; break;
                    case ISAAC_BAND_SH:   reject_right = SH_LR_RIGHT_REJ;break;
                    case ISAAC_BAND_SK:   reject_right = SK_LR_RIGHT_REJ;break;
                    default:
                        cpl_msg_warning(cpl_func, "unsupported filter");
                        reject_right = 0;
                        break;
                }
            }
        }
    } else if (isaac_spc_startrace_config.arm == 2) {
        /* LW */
        if (reject_left < 0) reject_left = 150;
        if (reject_right < 0) reject_right = 150;
    
    } else {
        cpl_msg_error(cpl_func, "Unrecognized mode - abort");
        return NULL;
    }

    /* Filter the image */
    kernel = cpl_mask_new(3, 3);
    cpl_mask_not(kernel);
    filtered = cpl_image_new(cpl_image_get_size_x(in), cpl_image_get_size_y(in),
                             cpl_image_get_type(in));
    if (cpl_image_filter_mask(filtered, in, kernel, CPL_FILTER_MEDIAN,
                              CPL_BORDER_FILTER)) {
        cpl_msg_error(cpl_func, "cannot filter the image");
        cpl_mask_delete(kernel);
        return NULL;
    }
    cpl_mask_delete(kernel);

    /* Fill the vector to fit */
    nsamples = cpl_image_get_size_x(in) - reject_right - reject_left;
    fit_x = cpl_matrix_new(1, nsamples);
    pfit_x = cpl_matrix_get_data(fit_x);
    fit_y = cpl_vector_new(nsamples);
    pfit_y = cpl_vector_get_data(fit_y);
    for (i=0; i < nsamples; i++) {
        pfit_x[i] = (double)(i+1+reject_left);
        pfit_y[i] = cpl_image_get_centroid_y_window(filtered, i+1+reject_left, 
                low_side, i+1+reject_left, up_side);
    }
    cpl_image_delete(filtered);
    
    /* Apply the fit */
    fitted = cpl_polynomial_new(1);
    if (cpl_polynomial_fit(fitted, fit_x, &sampsym, fit_y, NULL, CPL_FALSE,
                           NULL, &(isaac_spc_startrace_config.degree))) {
        cpl_msg_error(cpl_func, "cannot fit the shape");
        cpl_polynomial_delete(fitted);
        cpl_matrix_delete(fit_x);
        cpl_vector_delete(fit_y);
        return NULL;
    }

    /* Display if requested */
    if (isaac_spc_startrace_config.display) {
        cpl_vector * vfit_x
            = cpl_vector_wrap(nsamples, pfit_x);
        toplot = cpl_bivector_wrap_vectors(vfit_x, fit_y);
        cpl_plot_bivector(NULL, NULL, NULL, toplot);
        cpl_bivector_unwrap_vectors(toplot);
        (void)cpl_vector_unwrap(vfit_x);
    }

    cpl_vector_fill_polynomial_fit_residual(fit_y, fit_y, NULL, fitted,
                                            fit_x, NULL);
    *fit_quality = cpl_vector_product(fit_y, fit_y) / nsamples;

    /* Display if requested */
    if (isaac_spc_startrace_config.display) {
        cpl_vector * vfit_x
            = cpl_vector_wrap(nsamples, pfit_x);
        toplot = cpl_bivector_wrap_vectors(vfit_x, fit_y);
        for (i=0; i<nsamples; i++) {
            pfit_y[i] = cpl_polynomial_eval_1d(fitted, pfit_x[i], NULL);
        }
        cpl_plot_bivector(NULL, NULL, NULL, toplot);
        cpl_bivector_unwrap_vectors(toplot);
        (void)cpl_vector_unwrap(vfit_x);
    }

    /* Free and return */
    cpl_matrix_delete(fit_x);
    cpl_vector_delete(fit_y);
    return fitted;
}

/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    Compute the 2d distortion
  @param    shapes      the array with the polynomials
  @param    spec_pos    the spectra positions
  @param    nshapes     the number of polynomials
  @param    nx          the x size of the image
  @return   The 2d polynomial or NULL on error case
 */
/*----------------------------------------------------------------------------*/
static cpl_polynomial * isaac_spc_startrace_distor(
        cpl_polynomial  **  shapes,
        cpl_vector      *   spec_pos,
        int                 nshapes,
        int                 nx)
{
    cpl_polynomial      *   distor;
    cpl_matrix          *   xy_vec;
    double              *   x_data;
    double              *   y_data;
    cpl_vector          *   z_vec;
    double              *   z_data;
    int                     nb_xpoints, npoints;
    int                     pos;
    int                     i, j;
    const cpl_size          maxdeg = 2;

    /* Initialise */
    nb_xpoints = 20;
    npoints = nb_xpoints * nshapes;

    /* Prepare the surface */
    xy_vec = cpl_matrix_new(2, npoints);
    x_data = cpl_matrix_get_data(xy_vec);
    y_data = x_data + npoints;
    z_vec = cpl_vector_new(npoints);
    z_data = cpl_vector_get_data(z_vec);
    for (j=0; j<nshapes; j++) {
        for (i=0; i<nb_xpoints; i++) {
            pos = i + j*nb_xpoints;
            x_data[pos] = nx * (i+1) / nb_xpoints;
            y_data[pos] = cpl_vector_get(spec_pos, j);
            z_data[pos] = cpl_polynomial_eval_1d(shapes[j], x_data[pos], NULL);
        }
    }
    
    /* Apply the fit */
    distor = cpl_polynomial_new(2);
    if (cpl_polynomial_fit(distor, xy_vec, NULL, z_vec, NULL, CPL_FALSE, NULL,
                           &maxdeg)) {
        cpl_msg_error(cpl_func, "cannot fit the distortion");
        cpl_matrix_delete(xy_vec);
        cpl_vector_delete(z_vec);
        cpl_polynomial_delete(distor);
        return NULL;
    }
    cpl_matrix_delete(xy_vec);
    cpl_vector_delete(z_vec);

    return distor;
}

/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    Save the recipe products on disk
  @param    set_tot     the input frame set
  @param    tab         table to save
  @param    star_pos    the star Y-positions
  @param    spec_pos    the spectra positions
  @param    shapes      the spectra shapes
  @param    set         the frames from the current setting
  @param    resol       the resolution ('L' or 'M')
  @param    parlist     the input list of parameters
  @return   0 if everything is ok, -1 otherwise
 */
/*----------------------------------------------------------------------------*/
static
cpl_error_code isaac_spc_startrace_save(cpl_frameset            * set_tot,
                                        const cpl_table         * tab,
                                        const cpl_matrix        * star_pos,
                                        const cpl_vector        * spec_pos,
                                        cpl_polynomial         ** shapes,
                                        const cpl_frameset      * set,
                                        char                      resol,
                                        const cpl_parameterlist * parlist)
{
    cpl_propertylist    *   plist;
    cpl_propertylist    *   qclist;
    cpl_propertylist    *   paflist;
    const cpl_frame           *   ref_frame;
    char                    colname[128];
    cpl_table           *   shapes_table;
    cpl_table           *   pos_table;
    const char          *   procatt;
    const char          *   procat;
    char                *   filename;
    int                     i;

    /* Get the QC params in qclist */
    qclist = cpl_propertylist_new();

    /* Add QC parameters */
    cpl_propertylist_append_double(qclist, "ESO QC CORR_IS1",
            isaac_spc_startrace_config.corr_is1);
    cpl_propertylist_append_double(qclist, "ESO QC CORR_IS2",
            isaac_spc_startrace_config.corr_is2);
    cpl_propertylist_append_double(qclist, "ESO QC CORR_IS3",
            isaac_spc_startrace_config.corr_is3);
    cpl_propertylist_append_double(qclist, "ESO QC FITMSE",
            isaac_spc_startrace_config.shapes_fit_q);
    cpl_propertylist_append_double(qclist, "ESO QC DIST1",
            isaac_spc_startrace_config.dist1);
    cpl_propertylist_append_double(qclist, "ESO QC DISTX",
            isaac_spc_startrace_config.distx);
    cpl_propertylist_append_double(qclist, "ESO QC DISTY",
            isaac_spc_startrace_config.disty);
    cpl_propertylist_append_double(qclist, "ESO QC DISTXY",
            isaac_spc_startrace_config.distxy);
    cpl_propertylist_append_double(qclist, "ESO QC DISTXX",
            isaac_spc_startrace_config.distxx);
    cpl_propertylist_append_double(qclist, "ESO QC DISTYY",
            isaac_spc_startrace_config.distyy);
    
    /* Write the distortion table  */
    procatt = isaac_spc_startrace_config.arm == 1
        ? ISAAC_SPC_STARTRACE_SW_RES : ISAAC_SPC_STARTRACE_LW_RES;
    filename = cpl_sprintf("isaac_spc_startrace_2dpoly_%cR.fits", resol);
    irplib_dfs_save_table(set_tot,
            parlist,
            set,
            tab,
            NULL,
            "isaac_spc_startrace",
            procatt,
            qclist,
            NULL,
            PACKAGE "/" PACKAGE_VERSION,
            filename);
    cpl_free(filename);

    /* Save the positions table */
    pos_table = cpl_table_new(cpl_matrix_get_ncol(star_pos));

    /* star_pos is _not_ modified */
    cpl_table_wrap_double(pos_table, cpl_matrix_get_data((cpl_matrix*)star_pos),
                          "Star_positions");
    /* spec_pos is _not_ modified */
    cpl_table_wrap_double(pos_table, cpl_vector_get_data((cpl_vector*)spec_pos),
                          "Spec_positions");

    procat = isaac_spc_startrace_config.arm == 1
        ? ISAAC_SPC_STARTRACE_SW_POS : ISAAC_SPC_STARTRACE_LW_POS;
    filename = cpl_sprintf("isaac_spc_startrace_positions_%cR.fits", resol);
    irplib_dfs_save_table(set_tot,
            parlist,
            set,
            tab,
            NULL,
            "isaac_spc_startrace",
            procat,
            qclist,
            NULL,
            PACKAGE "/" PACKAGE_VERSION,
            filename);
    cpl_free(filename);

    (void)cpl_table_unwrap(pos_table, "Star_positions");
    (void)cpl_table_unwrap(pos_table, "Spec_positions");
    cpl_table_delete(pos_table);

    /* Save the shapes table */
    shapes_table = cpl_table_new(isaac_spc_startrace_config.degree + 1); 
    for (i=0; i<cpl_vector_get_size(spec_pos); i++) {
        sprintf(colname, "Spec_%d", i+1);
        cpl_table_new_column(shapes_table, colname, CPL_TYPE_DOUBLE);
    }
    for (i=0; i<cpl_vector_get_size(spec_pos); i++) {
        cpl_size j;
        sprintf(colname, "Spec_%d", i+1);
        for (j=0; j<isaac_spc_startrace_config.degree + 1; j++) {
            cpl_table_set_double(shapes_table, colname, j, 
                    cpl_polynomial_get_coeff(shapes[i], &j));
        }
    }
    procat = isaac_spc_startrace_config.arm == 1 ?
        ISAAC_SPC_STARTRACE_SW_SHAPE : ISAAC_SPC_STARTRACE_LW_SHAPE;
    filename = cpl_sprintf("isaac_spc_startrace_shapes_%cR.fits", resol);
    irplib_dfs_save_table(set_tot,
            parlist,
            set,
            tab,
            NULL,
            "isaac_spc_startrace",
            procat,
            qclist,
            NULL,
            PACKAGE "/" PACKAGE_VERSION,
            filename);
    cpl_free(filename);
    cpl_table_delete(shapes_table);

    /* Get the reference frame */
    ref_frame = irplib_frameset_get_first_from_group(set, CPL_FRAME_GROUP_RAW);

    /* Get FITS header from reference file */
    if ((plist=cpl_propertylist_load(cpl_frame_get_filename(ref_frame),
                    0)) == NULL) {
        cpl_msg_error(cpl_func, "getting header from reference frame");
        cpl_propertylist_delete(qclist);
        return CPL_ERROR_UNSPECIFIED;
    }
   
    /* Get the keywords for the paf file */
    paflist = cpl_propertylist_new();
    cpl_propertylist_copy_property_regexp(paflist, plist, 
        "^(ARCFILE|MJD-OBS|INSTRUME|ESO TPL ID|ESO TPL NEXP|ESO DPR CATG|"
        "ESO DPR TECH|ESO DPR TYPE|DATE-OBS|ESO INS GRAT NAME|"
        "ESO OCS SELECT-ARM)$", 0);
    cpl_propertylist_delete(plist);

    /* Copy the QC in paflist */
    cpl_propertylist_copy_property_regexp(paflist, qclist, ".", 0);
    cpl_propertylist_delete(qclist);

    /* PRO.CATG */
    cpl_propertylist_update_string(paflist, CPL_DFS_PRO_CATG, procatt);

    /* Save the PAF file */
    filename = cpl_sprintf("isaac_spc_startrace_%cR.paf", resol);
    cpl_dfs_save_paf("ISAAC",
            "isaac_spc_startrace",
            paflist,
            filename);
    cpl_free(filename);
    cpl_propertylist_delete(paflist);
    return  0;
}

